import { TeamSpeak, QueryProtocol } from "ts3-nodejs-library";
import { GraphQLClient, gql } from "graphql-request";

type Variables = { [key: string]: any };

var config = require("./config.json");

var voter = "";
const teamspeak = new TeamSpeak({
    host: config.host,
    protocol: QueryProtocol.RAW, //optional
    queryport: 10011, //optional
    serverport: 9987,
    username: config.username,
    password: config.password,
    nickname: "10Man",
});

let commands = new Map([
    ["!hello", "Say hello"],
    ["!elo", "Show your elo"],
    [
        "!register <STEAMID64>",
        "Register with your STEAMID64 (https://steamid.io/)",
    ],
    ["!play", "Go into the queue"],
    ["!cancel", "Cancel you queue"],
    ["!help", "Show this help text"],
    [
        "!ban <MAP>",
        "Ban <MAP> if you are a captain and it is your turn to vote",
    ],
]);

let admin_commands = new Map([
    ["!cancelMatch", "Cancel the ongoing match"],
    ["!finish <WINNER>", "Finish the ongoing match with WINNER (TEAM1, TEAM2)"],
    ["!set <STEAMID64> <ATTRIBUTE> <VALUE>", "Set ATTRIBUTE (elo) to VALUE"],
    ["!move <DIRECTION>", "Move clients in DIRECTION (optional, back)"],
]);

teamspeak.on("ready", async () => {
    const whoami = await teamspeak.whoami();
    const channel = await teamspeak.channelFind("10 Man");
    teamspeak.clientMove(whoami.clientId, channel[0].cid);
});

teamspeak.on("close", async () => {
    console.log("disconnected, trying to reconnect...");
    await teamspeak.reconnect(-1, 60000);
    console.log("reconnected!");
});

teamspeak.on("error", (e) => {
    console.log("Caught an error!");
    console.error(e);
});

teamspeak.on("textmessage", (ev) => {
    const argv = ev.msg.split(" ");
    const clid = ev.invoker.clid;
    const teamspeakID = ev.invoker.uniqueIdentifier;

    switch (argv[0]) {
        case "!hello":
            message(clid, "Hello " + ev.invoker.nickname);
            break;
        case "!elo":
            elo(clid, teamspeakID);
            break;
        case "!register":
            register(clid, teamspeakID, argv[1]);
            break;
        case "!play":
            play(clid, teamspeakID);
            break;
        case "!cancel":
            cancel(clid, teamspeakID);
            break;
        case "!cancelMatch":
            cancel_match(clid, teamspeakID);
            break;
        case "!finish":
            finish_match(clid, teamspeakID, argv[1]);
            break;
        case "!set":
            set(clid, teamspeakID, argv[1], argv[2], argv[3]);
            break;
        case "!help":
            help(clid, teamspeakID);
            break;
        case "!move":
            move_clients_wrapper(clid, teamspeakID, argv[1]);
            break;
        case "!ban":
            ban(clid, teamspeakID, argv[1]);
            break;
    }
});

function message(clid: string, msg: string) {
    teamspeak.sendTextMessage(clid, 1, msg);
}

function message_channel(msg: string) {
    teamspeak.channelFind("10 Man").then((channel) => {
        teamspeak.sendTextMessage(channel[0].cid, 2, msg);
    });
}

// wrapper to send msg to a teamspeakID
async function message_id(teamspeakID: string, msg: string) {
    const client = await teamspeak
        .getClientByUid(teamspeakID)
        .then((client) => {
            message(client!.clid, msg);
        })
        .catch(function (e) {
            console.error(e);
        });
}

function get_error(e: unknown): string {
    try {
        return JSON.parse(JSON.stringify(e)).response.errors[0].message;
    } catch {
        return e as string;
    }
}

async function request(query: string, variables?: Variables): Promise<any> {
    const endpoint = config.endpoint!;
    const graphQLClient = new GraphQLClient(endpoint, {
        headers: {},
    });

    const data = await graphQLClient.request(query, variables);
    return data;
}

async function get_user_id(clid: string, steamID64: string): Promise<bigint> {
    const query = gql`
        query ($steamID: String!) {
            userBySteam(steamID: $steamID) {
                id
            }
        }
    `;
    const variables = { steamID: steamID64 };

    try {
        const data = await request(query, variables);
        const id = data.userBySteam.id;
        return id;
    } catch (e) {
        const error = get_error(e);
        switch (error) {
            case "no result":
                message(clid, "This user is not registered.");
                break;
            default:
                console.error(error);
                break;
        }
        return -1n;
    }
}

async function is_admin(clid: string, teamspeakID: string): Promise<boolean> {
    const query = gql`
        query ($teamspeakID: String!) {
            userByTS(teamspeakID: $teamspeakID) {
                admin
            }
        }
    `;
    const variables = { teamspeakID: teamspeakID };

    try {
        const data = await request(query, variables);
        const is_admin = data.userByTS.admin;
        return is_admin;
    } catch (e) {
        const error = get_error(e);
        switch (error) {
            case "no result":
                message(clid, "You are not registered with this teamspeakID.");
                break;
            default:
                console.error(error);
                break;
        }
        return false;
    }
}

async function elo(clid: string, teamspeakID: string) {
    const query = gql`
        query ($teamspeakID: String!) {
            userByTS(teamspeakID: $teamspeakID) {
                elo
            }
        }
    `;
    const variables = { teamspeakID: teamspeakID };

    try {
        const data = await request(query, variables);
        message(clid, "Elo: " + data.userByTS.elo);
    } catch (e) {
        const error = get_error(e);
        switch (error) {
            case "no result":
                message(clid, "You are not registered with this teamspeakID.");
                break;
            default:
                console.error(error);
                break;
        }
    }
}

async function register(clid: string, teamspeakID: string, steamID: string) {
    const mutation = gql`
        mutation ($steamID: String!, $teamspeakID: String!) {
            createUser(input: { steamID: $steamID, teamspeakID: $teamspeakID })
        }
    `;
    const variables = { teamspeakID: teamspeakID, steamID: steamID };

    try {
        const data = await request(mutation, variables);
        message(clid, "Registered successfully.");
    } catch (e) {
        const error = get_error(e);
        switch (error) {
            case "not unique":
                message(
                    clid,
                    "You are already registered with this teamspeakID or steamID. Please contact an administrator if you believe that this is an error."
                );
                break;
            case "wrong steamID64":
                message(clid, "You entered a wrong steamID64.");
                break;
            case "must be defined":
                message(clid, "Usage: !register STEAMID64");
                break;
            default:
                console.error(error);
                break;
        }
    }
}

async function play(clid: string, teamspeakID: string) {
    const mutation = gql`
        mutation ($teamspeakID: String!) {
            startQueue(teamspeakID: $teamspeakID)
        }
    `;
    const variables = { teamspeakID: teamspeakID };

    try {
        const data = await request(mutation, variables);
        const players = data.startQueue;
        message_channel(players + " player(s) in queue.");

        if (players == 10) {
            start_match();
        }
    } catch (e) {
        const error = get_error(e);
        switch (error) {
            case "already in queue":
                message(clid, "Already in queue.");
                break;
            case "queue is full":
                message(clid, "Queue is full.");
                break;
            case "cannot queue, match still ongoing":
                message(clid, "Cannot queue, a match is ongoing.");
                break;
            case "no result":
                message(clid, "You are not registered with this teamspeakID.");
                break;
            default:
                console.error(error);
                break;
        }
    }
}

async function cancel(clid: string, teamspeakID: string) {
    const mutation = gql`
        mutation ($teamspeakID: String!) {
            cancelQueue(teamspeakID: $teamspeakID)
        }
    `;
    const variables = { teamspeakID: teamspeakID };

    try {
        const data = await request(mutation, variables);
        message_channel(data.cancelQueue + " player(s) in queue.");
    } catch (e) {
        const error = get_error(e);
        switch (error) {
            case "user not in queue":
                message(clid, "You are not in queue.");
                break;
            default:
                console.error(error);
                break;
        }
    }
}

async function cancel_match(clid: string, teamspeakID: string) {
    const admin = await is_admin(clid, teamspeakID);
    if (!admin) {
        message(clid, "Only an admin can do that.");
        return;
    }

    const mutation = gql`
        mutation {
            cancelMatch
        }
    `;

    try {
        const data = await request(mutation);
        message_channel("Match cancelled.");
    } catch (e) {
        const error = get_error(e);
        switch (error) {
            case "no match ongoing":
                message(clid, "No ongoing match.");
                break;
            default:
                console.error(error);
                break;
        }
    }
}

async function finish_match(clid: string, teamspeakID: string, winner: string) {
    if (winner !== "TEAM1" && winner !== "TEAM2") {
        message(clid, "Illegal winning team. Only 'TEAM1' or 'TEAM2' allowed.");
        return;
    }

    const admin = await is_admin(clid, teamspeakID);
    if (!admin) {
        message(clid, "Only an admin can do that.");
        return;
    }

    // move clients back to main channel
    move_clients(true);

    const mutation = gql`
        mutation ($winner: String!) {
            finishMatch(winner: $winner)
        }
    `;
    const variables = { winner: winner };

    try {
        const data = await request(mutation, variables);
        message_channel("Match finished.");
    } catch (e) {
        const error = get_error(e);
        switch (error) {
            case "no match ongoing":
                message(clid, "No ongoing match.");
                break;
            default:
                console.error(error);
                break;
        }
    }
}

async function start_match() {
    const mutation = gql`
        mutation {
            createMatch
        }
    `;

    try {
        const data = await request(mutation);
        message_channel("Started match.");
        move_clients();
        vote_map();
    } catch (e) {
        const error = get_error(e);
        switch (error) {
            default:
                console.error(error);
                break;
        }
    }
}

// move clients manually, in case match was started from website
async function move_clients_wrapper(
    clid: string,
    teamspeakID: string,
    direction: string
) {
    const admin = await is_admin(clid, teamspeakID);
    if (!admin) {
        message(clid, "Only an admin can do that.");
        return;
    }

    if (direction === "back") {
        move_clients(true);
    } else {
        move_clients(false);
    }
}

async function move_clients(back: boolean = false) {
    const query = gql`
        query {
            getTeams {
                team1 {
                    teamspeakID
                }
                team2 {
                    teamspeakID
                }
            }
        }
    `;

    try {
        const data = await request(query);
        const team1 = data.getTeams.team1;
        const team2 = data.getTeams.team2;
        var channel = await teamspeak.channelFind("10 Man");

        if (!back) {
            channel = await teamspeak.channelFind("Team 1");
        }

        for (let i of team1) {
            const client = await teamspeak
                .getClientByUid(i.teamspeakID)
                .then((client) => {
                    teamspeak
                        .clientMove(client!.clid, channel[0].cid)
                        .catch(function (e) {
                            console.error(e);
                        });
                })
                .catch(function (e) {
                    console.error(e);
                });
        }

        if (!back) {
            channel = await teamspeak.channelFind("Team 2");
        }
        for (let i of team2) {
            const client = await teamspeak
                .getClientByUid(i.teamspeakID)
                .then((client) => {
                    teamspeak
                        .clientMove(client!.clid, channel[0].cid)
                        .catch(function (e) {
                            console.error(e);
                        });
                })
                .catch(function (e) {
                    console.error(e);
                });
        }
    } catch (e) {
        const error = get_error(e);
        switch (error) {
            default:
                console.error(error);
                break;
        }
    }
}

async function help(clid: string, teamspeakID: string) {
    const admin = await is_admin(clid, teamspeakID);
    var msg = "";

    if (admin) {
        msg += "\nAdmin only commands:\n";
        for (let entry of admin_commands) {
            msg += entry[0] + ": " + entry[1] + "\n";
        }
    }

    msg += "\nUser commands:\n";
    for (let entry of commands) {
        msg += entry[0] + ": " + entry[1] + "\n";
    }

    teamspeak.sendTextMessage(clid, 1, msg);
}

async function set(
    clid: string,
    teamspeakID: string,
    steamID64: string,
    attribute: string,
    value: string
) {
    const admin = await is_admin(clid, teamspeakID);
    if (!admin) {
        message(clid, "Only an admin can do that.");
        return;
    }

    switch (attribute) {
        case "elo":
            set_elo(clid, steamID64, value);
            break;
        default:
            message(clid, "Attribute not found.");
            break;
    }
}

async function set_elo(clid: string, steamID64: string, elo: string) {
    const id = await get_user_id(clid, steamID64);
    if (id == -1n) {
        return;
    }

    const mutation = gql`
        mutation ($id: Int!, $elo: Int!) {
            updateUser(id: $id, input: { elo: $elo }) {
                elo
            }
        }
    `;

    try {
        // converting can fail, so put it in the try block
        const variables = { id: id, elo: Number(BigInt(elo)) };
        const data = await request(mutation, variables);
        message(clid, "Elo updated successfully.");
    } catch (e) {
        console.error(e);
        message(clid, "Elo update failed.");
    }
}

async function vote_map() {
    var voting_side = "";

    // get captains
    const query = gql`
        query {
            getTeams {
                captain1 {
                    teamspeakID
                }
                captain2 {
                    teamspeakID
                }
            }
        }
    `;

    var data = await request(query);
    const captain1 = data.getTeams.captain1.teamspeakID;
    const captain2 = data.getTeams.captain2.teamspeakID;

    while (true) {
        // get voting information
        const mutation = gql`
            mutation {
                vetoMap {
                    status
                    maps
                }
            }
        `;

        var data = await request(mutation);
        var status = data.vetoMap.status;
        const maps = data.vetoMap.maps.join(", ");

        if (status === "DONE") {
            // Forbid votes
            voter = "";
            const msg = maps + " was selected.";
            message_id(captain1, msg);
            message_id(captain2, msg);
            break;
        }
        if (voting_side !== status) {
            voting_side = status;
            const msg =
                "Please ban one of the following maps with !ban: " + maps;

            if (voting_side === "TEAM1") {
                voter = captain1;
                message_id(captain1, msg);
            }

            if (voting_side === "TEAM2") {
                voter = captain2;
                message_id(captain2, msg);
            }
        }

        await sleep(10000);
    }
}

async function ban(clid: string, teamspeakID: string, map: string) {
    if (teamspeakID !== voter) {
        message(clid, "You are not allowed to ban right now");
        return;
    }

    const mutation = gql`
        mutation ($map: String!) {
            vetoMap(map: $map) {
                status
            }
        }
    `;
    const variables = { map: map };

    try {
        const data = await request(mutation, variables);
    } catch (e) {
        const error = get_error(e);
        switch (error) {
            case "no legal vote":
                message(clid, "You have not entered a correct mapname.");
                break;
            case "must be defined":
                message(clid, "Put a mapname after !ban.");
                break;
            default:
                message(
                    clid,
                    "Ban was not successful, please contact an admin."
                );
                console.error(error);
                break;
        }
    }
}

function sleep(ms: number) {
    return new Promise((resolve) => setTimeout(resolve, ms));
}
